/*
Project: chipKIT Uno32 Pulse Meter
Description: Uses Easy Pulse V1.1 sensor to sense PPG signal
from a finger tip. The project uses chipKIT Basic I/O shield to 
display the pulse waveform and pulse rate on the OLED screen.

Author: Rajendra Bhatt (rajbex@gmail.com)
Date: Jan 8, 2014
License: Creative Commons Attribution-ShareAlike 3.0.

*/


#include <IOShieldOled.h>
#include <Wprogram.h>
#define  OLED  IOShieldOledClass

#define cbOledChar      8                //number of bytes in a glyph
#define cbOledFontUser  (32*cbOledChar)  //number of bytes in user font table

#define Sampling_Time 5
#define Num_Samples 600
#define Peak_Threshold_Factor 75
#define Minimum_Range 50
#define Minimum_Peak_Separation 50  // 50*5=250 ms
#define DC_Added 10;
#define Samples_to_Display 400
#define Display_Sampling 5
#define x_axis_length 80
#define y_axis_length 31
#define Moving_Average_Num 5

int ADC_Samples[Num_Samples], Index1, Index2, Index3, Peak1, Peak2, Peak3;
long Pulse_Rate, Temp1=1L;
int Peak_Magnitude, Peak_Threshold, Minima, Range;
/* ------------------------------------------------------------ */
/*		Global Variables				*/
/* ------------------------------------------------------------ */

OLED  oled;

void
setup() {
  char  ch;  

  /* Initialize the OLED display before access.
  */
  oled.begin();
  Serial.begin(9600);
  
}

void loop() {
  Read_ADC_Samples();
  Serial.println("Sample Read Finished ");
  Remove_DC();
  Serial.println("DC component subtracted ");
  Scale_Data();
  Serial.println("Data scaled ");
  if (Range > Minimum_Range){  // ADC range is > 50, otherwise increse gain
    Filter_Data();
    Serial.println("Data Filtered ");
    Compute_Pulse_Rate();
    Serial.println("Pulse rate computed ");
    Display_Samples();
    Serial.println("Sample displayed ");
  
    // Now print heart rate
    oled.setCursor(12, 1);
    char strheart[4];
    //dtostrf(Pulse_Rate, 5,1, strheart);
    itoa(Pulse_Rate, strheart, 10);
    oled.putString(strheart);
    oled.setCursor(11, 3);
    oled.putString("BPM");
  }
  
  else{
    oled.clear();
    oled.setCursor(0, 1);
    oled.putString("No Pulse Signal!");
    oled.setCursor(0, 3);
    oled.putString("Check Sensor!");
  }
}

void Read_ADC_Samples(){
  for (int i = 0; i < Num_Samples; i++){
     //ADC_Samples[i] = 1023-analogRead(A1);
     ADC_Samples[i] = analogRead(A1);
     //delay_x(5);
     delayMicroseconds(Sampling_Time*1000);  
  }  
}

void Remove_DC(){
  Find_Minima(0);
  Serial.print("Minima = ");
  Serial.println(Minima);
  // Subtract DC (minima) 
  for (int i = 0; i < Num_Samples; i++){
     ADC_Samples[i] = ADC_Samples[i] - Minima;
  }
  Minima = 0;  // New minima is zero  
}

void Scale_Data(){
  // Find peak value
  Find_Peak(0);
  Serial.print("Peak = ");
  Serial.println(Peak_Magnitude);
  Range = Peak_Magnitude - Minima;
  Serial.print("Range = ");
  Serial.println(Range);
  // Sclae from 1 to 1023 
  if(Range > Minimum_Range){
    for (int i = 0; i < Num_Samples; i++){
       ADC_Samples[i] = 1 + ((ADC_Samples[i]-Minima)*1022)/Range;
     
    }
    Find_Peak(0);
    Find_Minima(0);
    Serial.print("Scaled Peak = ");
    Serial.println(Peak_Magnitude);
    Serial.print("Scaled Minima = ");
    Serial.println(Minima);
  }
}


void Filter_Data(){
  int Num_Points = 2*Moving_Average_Num+1;
  for (int i = Moving_Average_Num; i < Num_Samples-Moving_Average_Num; i++){
    int Sum_Points = 0;
    for(int k =0; k < Num_Points; k++){   
      Sum_Points = Sum_Points + ADC_Samples[i-Moving_Average_Num+k]; 
    }    
    ADC_Samples[i] = Sum_Points/Num_Points; 
  } 
}


void Compute_Pulse_Rate(){
  // Detect Peak magnitude and minima
  Find_Peak(Moving_Average_Num);
  Find_Minima(Moving_Average_Num);
  Range = Peak_Magnitude - Minima;
  Peak_Threshold = Peak_Magnitude*Peak_Threshold_Factor;
  Peak_Threshold = Peak_Threshold/100;
 
  // Now detect three peaks 
  Peak1 = 0;
  Peak2 = 0;
  Peak3 = 0;
  Index1 = 0;
  Index2 = 0;
  Index3 = 0;
  // Find first peak
  for (int j = Moving_Average_Num; j < Num_Samples-Moving_Average_Num; j++){
      if(ADC_Samples[j] >= ADC_Samples[j-1] && ADC_Samples[j] > ADC_Samples[j+1] && 
         ADC_Samples[j] > Peak_Threshold && Peak1 == 0){
        Peak1 = ADC_Samples[j];
        Index1 = j; 
      }
      
      // Search for second peak which is at least 10 sample time far
      if(Peak1 > 0 && j > (Index1+Minimum_Peak_Separation) && Peak2 == 0){
         if(ADC_Samples[j] >= ADC_Samples[j-1] && ADC_Samples[j] > ADC_Samples[j+1] && 
         ADC_Samples[j] > Peak_Threshold){
         Peak2 = ADC_Samples[j];
         Index2 = j; 
         } 
      } // Peak1 > 0
      
      // Search for the third peak which is at least 10 sample time far
      if(Peak2 > 0 && j > (Index2+Minimum_Peak_Separation) && Peak3 == 0){
         if(ADC_Samples[j] >= ADC_Samples[j-1] && ADC_Samples[j] > ADC_Samples[j+1] && 
         ADC_Samples[j] > Peak_Threshold){
            Peak3 = ADC_Samples[j];
            Index3 = j; 
         } 
      } // Peak2 > 0
    
  }
  Serial.print("Index1 = ");
  Serial.println(Index1);
  Serial.print("Index2 = ");
  Serial.println(Index2);
  Serial.print("Index3 = ");
  Serial.println(Index3);
  
  
  PR1 = (Index2-Index1)*Sampling_Time; // In milliseconds
  PR2 = (Index3-Index2)*Sampling_Time;
  Serial.print("PR1 = ");
  Serial.println(PR1);
  Serial.print("PR2 = ");
  Serial.println(PR2);
  if(PR1 > 0 && PR2 > 0){
    //float PR = (PR1+PR2)/2.0;
    //Serial.print("PR = ");
    //Serial.println(PR);
    //Pulse_Rate = PR;
    Pulse_Rate = (PR1+PR2)/2;
    Pulse_Rate = 60000/Pulse_Rate; // In BPM
     
  }
}

void Display_Samples(){
  oled.clear();
  oled.moveTo(0, 0);
  oled.drawLine(0,y_axis_length);
  oled.drawLine(x_axis_length,y_axis_length);
  
  for(int i = Moving_Average_Num; i < Samples_to_Display; i+=Display_Sampling){
    int y_pos = 1 + ((ADC_Samples[i]-Minima)*29)/Range;
    y_pos = 31-y_pos;
    oled.moveTo(i/Display_Sampling, y_pos);
    oled.drawPixel();
  }
  
}

void Find_Peak(int Num){
  Peak_Magnitude = 0;
  for (int m = Num; m < Num_Samples-Num; m++){
      if(Peak_Magnitude < ADC_Samples[m]){
        Peak_Magnitude = ADC_Samples[m];
     }
  }
}

void Find_Minima(int Num){
  Minima = 1024;
  for (int m = Num; m < Num_Samples-Num; m++){
      if(Minima > ADC_Samples[m]){
         Minima = ADC_Samples[m];
      }
  }
}


void delay_x(uint32_t millis_delay)
{
  uint16_t micros_now = (uint16_t)micros();

  while (millis_delay > 0) {
    if (((uint16_t)micros() - micros_now) >= 1000) {
      millis_delay--;
      micros_now += 1000;
    }
  }  
}
